class DataType:
    """ A class to identify what data you want to retrieve. This class has two parameters which are very useful.
        var:
            - Almost always a string representation of the data you want to retrieve. This project uses the types from:
                http://dderevjanik.github.io/agescx/formatscx/#format
                s: Signed integer       All integers (signed and unsigned) are parsed as little endian
                u: Unsigned integer     All integers (signed and unsigned) are parsed as little endian
                f: float
                c: Character string of fixed length
                (Empty): Is interpreted as regular byte data. In this project the '' is converted to 'data'
                str: Character string of variable length.
                    This type will read the number in bits given and parse it as an int. The number retrieved from it
                    will be the amount of bytes read as a character string.
            - Another option for the var parameter is to give a Struct (not Python struct) subclass as var. This will
            parse all DataType values in the Struct subclass. This can be handy for when blocks of data are repeated.
            - Per data type you can save how large the value is bit/byte wise. While this may be confusing not all
            values are in bit format. Some are in byte format.
                Bit format: [s, u, f, str]
                Byte format: [c, data]
            - To define a length to your data type, write the datatype with the length behind it (no whitespaces)
                Example 1:  s16             > A signed integer of 16 bits.
                Example 2:  f32             > A 32 bit floating point number.
                Example 3:  c4              > A 32 bit (4 bytes) character string.
                Example 4:  str16           > A 16 bit integer will be parsed (n). Now n bytes will be read as character
                                            string.
                Example 5:  TerrainStruct   > The TerrainStruct will be instantiated and DataTypes from that struct will
                                            be loaded in it's place.
        repeat:
            The amount of times the above datatype needs to be repeated
    """

    __slots__ = [
        'var',
        '_repeat',
        'log_value',
        'type',
        'length',
        '_debug_retriever_name'
    ]

    _debug_retriever_name: str

    def __init__(self, var="0", repeat=1, log_value=False, type_length=None):
        self.var = var
        self._repeat = repeat
        self.log_value = log_value

        if type_length is not None:
            self.type, self.length = type_length
        else:
            self.type, self.length = datatype_to_type_length(self.var)

    def get_struct_name(self) -> str:
        if self.var.startswith("struct:"):
            return self.var[7:]
        raise ValueError("Datatype var is not a struct")

    @property
    def type_and_length(self):
        return self.type, self.length

    @property
    def repeat(self):
        return self._repeat

    @repeat.setter
    def repeat(self, value):
        if self.log_value:
            print(f"[DataType] {self._debug_retriever_name} 'repeat' set to {value} (was: {self._repeat})")
        self._repeat = value

    def to_simple_string(self):
        return f"{self._repeat} * {self.var}"

    def __repr__(self):
        return f"[DataType] " + self.to_simple_string()

    def duplicate(self):
        return DataType(
            var=self.var,
            repeat=self.repeat,
            log_value=self.log_value,
            type_length=(self.type, self.length)
        )


def datatype_to_type_length(var):
    """Returns the type and length of a datatype. So: 'int32' returns 'int', 32. """
    if var[:7] == "struct:":
        return "struct", 0

    # Filter numbers out for length, filter text for type
    var_len = int(''.join(filter(str.isnumeric, var)))
    var_type = ''.join(filter(str.isalpha, var))

    if var_type == '':
        var_type = "data"

    if var_type not in datatype_types:
        raise ValueError(f"Unknown variable type '{var_type}'")

    # Divide by 8, and parse from float to int
    if var_type not in ["c", "data"]:
        var_len = int(var_len / 8)

    return var_type, var_len


datatype_types = [
    "s",  # Signed int
    "u",  # Unsigned int
    "f",  # FloatingPoint
    "c",  # Character string
    "str",  # Variable length string
    "data",  # Data (Can be changed by used using bytes_to_x functions)
]



from typing import List

from AoE2ScenarioParser.objects.data_objects.terrain_tile import TerrainTile
from AoE2ScenarioParser.objects.managers.map_manager import MapManager
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.support import Support


class MapManagerDE(MapManager):
    _link_list = [
        RetrieverObjectLink("map_color_mood", "Map", "map_color_mood"),
        RetrieverObjectLink("collide_and_correct", "Map", "collide_and_correct"),
        RetrieverObjectLink("villager_force_drop", "Map", "villager_force_drop", Support(since=1.37)),
        RetrieverObjectLink("map_width", "Map", "map_width"),
        RetrieverObjectLink("map_height", "Map", "map_height"),
        RetrieverObjectLink("terrain", "Map", "terrain_data", process_as_object=TerrainTile),
    ]

    def __init__(self,
                 map_color_mood: str,
                 collide_and_correct: bool,
                 villager_force_drop: bool,
                 map_width: int,
                 map_height: int,
                 terrain: List[TerrainTile],
                 **kwargs,
                 ):
        self.map_color_mood = map_color_mood
        self.collide_and_correct = collide_and_correct
        self.villager_force_drop = villager_force_drop

        super().__init__(map_width, map_height, terrain, **kwargs)

    @property
    def script_name(self):
        raise DeprecationWarning("The attribute script_name is handled through the xs_manager. "
                                 "scenario.xs_manager.script_name")

    @script_name.setter
    def script_name(self, val):
        raise DeprecationWarning("The attribute script_name is handled through the xs_manager. "
                                 "scenario.xs_manager.script_name")



from typing import List

from AoE2ScenarioParser.datasets.players import ColorId
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class PlayerDataThree(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("initial_camera_x", "Units", "player_data_3[__index__].initial_camera_x"),
        RetrieverObjectLink("initial_camera_y", "Units", "player_data_3[__index__].initial_camera_y"),

        # Sort of duplicates. Values gets overridden by other diplomacy values, these just get changed based on that
        RetrieverObjectLink("diplomacy_for_interaction", "Units", "player_data_3[__index__].diplomacy_for_interaction"),
        RetrieverObjectLink("diplomacy_for_ai_system", "Units", "player_data_3[__index__].diplomacy_for_ai_system"),

        # Duplicates, only here so they're replaced in the scenario file. In case it impacts gameplay
        RetrieverObjectLink("aok_allied_victory", "Units", "player_data_3[__index__].aok_allied_victory"),
        RetrieverObjectLink("color", "Units", "player_data_3[__index__].color"),

        # AFAIK - These are stored but have no impact on the game, last camera xy position with player selected
        # RetrieverObjectLink("editor_camera_x", "Units", "player_data_3[__index__].editor_camera_x"),
        # RetrieverObjectLink("editor_camera_y", "Units", "player_data_3[__index__].editor_camera_y"),
    ]

    def __init__(
            self,
            initial_camera_x: int,
            initial_camera_y: int,
            aok_allied_victory: int,
            diplomacy_for_interaction: List[int],
            diplomacy_for_ai_system: List[int],
            color: int,
            **kwargs):
        super().__init__(**kwargs)

        self.initial_camera_x = initial_camera_x
        self.initial_camera_y = initial_camera_y
        self.aok_allied_victory = aok_allied_victory
        self.diplomacy_for_interaction = diplomacy_for_interaction
        self.diplomacy_for_ai_system = diplomacy_for_ai_system
        self.color: ColorId = ColorId(color)



from __future__ import annotations

import math
from typing import Union

from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.helper import raise_if_not_int_subclass
from AoE2ScenarioParser.helper.pretty_format import pretty_format_name
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.support.tile import Tile
from AoE2ScenarioParser.scenarios.scenario_store import actions
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class Unit(AoE2Object):
    _link_list = [
        RetrieverObjectLink("player", retrieve_history_number=0),
        RetrieverObjectLink("x", "Units", "players_units[__index__].units[__index__].x"),
        RetrieverObjectLink("y", "Units", "players_units[__index__].units[__index__].y"),
        RetrieverObjectLink("z", "Units", "players_units[__index__].units[__index__].z"),
        RetrieverObjectLink("reference_id", "Units", "players_units[__index__].units[__index__].reference_id"),
        RetrieverObjectLink("unit_const", "Units", "players_units[__index__].units[__index__].unit_const"),
        RetrieverObjectLink("status", "Units", "players_units[__index__].units[__index__].status"),
        RetrieverObjectLink("rotation", "Units", "players_units[__index__].units[__index__].rotation"),
        RetrieverObjectLink("initial_animation_frame", "Units",
                            "players_units[__index__].units[__index__].initial_animation_frame"),
        RetrieverObjectLink("garrisoned_in_id", "Units", "players_units[__index__].units[__index__].garrisoned_in_id"),
    ]

    def __init__(self,
                 player: Union[int, PlayerId],
                 x: float,
                 y: float,
                 z: float,
                 reference_id: int,
                 unit_const: int,
                 status: int,
                 rotation: float,
                 initial_animation_frame: int,
                 garrisoned_in_id: int,
                 **kwargs
                 ):
        raise_if_not_int_subclass([unit_const])

        super().__init__(**kwargs)

        self._player: PlayerId = PlayerId(player)
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.reference_id: int = reference_id
        self.unit_const: int = unit_const
        self.status: int = status
        self.rotation: float = rotation
        self.initial_animation_frame: int = initial_animation_frame
        self.garrisoned_in_id: int = garrisoned_in_id

    @property
    def player(self) -> PlayerId:
        return self._player

    @player.setter
    def player(self, player: Union[int, PlayerId]):
        actions.unit_change_ownership(self._host_uuid, player, self)
        self._player = player

    @property
    def tile(self) -> Tile:
        return Tile(math.floor(self.x), math.floor(self.y))
        # Floor x and y as location (0.9, 0.9) is still Tile[x=0, y=0]

    @tile.setter
    def tile(self, tile: Tile) -> None:
        self.x = tile.x
        self.y = tile.y

    @property
    def name(self) -> str:
        unit_enum = helper.get_enum_from_unit_const(self.unit_const)
        if unit_enum:
            return pretty_format_name(unit_enum.name)
        else:
            return f"Unknown{self.unit_const}"  # e.g."Unknown411"



from typing import Optional, Tuple

from AoE2ScenarioParser.datasets.terrains import TerrainId
from AoE2ScenarioParser.helper.helper import i_to_xy
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class TerrainTile(AoE2Object):
    """Object for handling a tile in the map."""

    _link_list = [
        RetrieverObjectLink("terrain_id", "Map", "terrain_data[__index__].terrain_id"),
        RetrieverObjectLink("elevation", "Map", "terrain_data[__index__].elevation"),
        RetrieverObjectLink("layer", "Map", "terrain_data[__index__].layer"),
        RetrieverObjectLink("_index", retrieve_history_number=0),
    ]

    def __init__(self, terrain_id: int = TerrainId.GRASS_1, elevation: int = 0, layer: int = -1, _index: int = - 1,
                 **kwargs):
        self.terrain_id = terrain_id
        self.elevation = elevation
        self.layer = layer
        self._index = _index
        self._xy: Optional[Tuple[int, int]] = None

        super().__init__(**kwargs)

    @property
    def x(self) -> int:
        return self.xy[0]

    @property
    def y(self) -> int:
        return self.xy[1]

    @property
    def i(self) -> int:
        return self._index

    @property
    def xy(self) -> Tuple[int, int]:
        if not self._xy:
            self._xy = i_to_xy(self._index, getters.get_map_size(self._host_uuid))
        return self._xy


def reset_terrain_index(tile, new_index):
    tile._index = new_index
    tile._xy = None



# This file was generated based on: versions/DE/v1.41/conditions.json
from typing import Union

from AoE2ScenarioParser.datasets.conditions import ConditionId
from AoE2ScenarioParser.objects.data_objects.condition import Condition


class NewConditionSupport:
    def __init__(self, trigger_ref):
        self._trigger_ref = trigger_ref

    def none(
            self,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.NONE,
        )

    def bring_object_to_area(
            self,
            unit_object: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.BRING_OBJECT_TO_AREA,
            unit_object=unit_object,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            inverted=inverted,
        )

    def bring_object_to_object(
            self,
            unit_object: Union[int, None] = None,
            next_object: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.BRING_OBJECT_TO_OBJECT,
            unit_object=unit_object,
            next_object=next_object,
            inverted=inverted,
        )

    def own_objects(
            self,
            quantity: Union[int, None] = None,
            object_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OWN_OBJECTS,
            quantity=quantity,
            object_list=object_list,
            source_player=source_player,
            object_group=object_group,
            object_type=object_type,
        )

    def own_fewer_objects(
            self,
            quantity: Union[int, None] = None,
            object_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OWN_FEWER_OBJECTS,
            quantity=quantity,
            object_list=object_list,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
        )

    def objects_in_area(
            self,
            quantity: Union[int, None] = None,
            object_list: Union[int, None] = None,
            source_player: Union[int, None] = None,
            area_x1: Union[int, None] = None,
            area_y1: Union[int, None] = None,
            area_x2: Union[int, None] = None,
            area_y2: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            inverted: Union[int, None] = None,
            object_state: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECTS_IN_AREA,
            quantity=quantity,
            object_list=object_list,
            source_player=source_player,
            area_x1=area_x1,
            area_y1=area_y1,
            area_x2=area_x2,
            area_y2=area_y2,
            object_group=object_group,
            object_type=object_type,
            inverted=inverted,
            object_state=object_state,
        )

    def destroy_object(
            self,
            unit_object: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.DESTROY_OBJECT,
            unit_object=unit_object,
            inverted=inverted,
        )

    def capture_object(
            self,
            unit_object: Union[int, None] = None,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.CAPTURE_OBJECT,
            unit_object=unit_object,
            source_player=source_player,
            inverted=inverted,
        )

    def accumulate_attribute(
            self,
            quantity: Union[int, None] = None,
            attribute: Union[int, None] = None,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.ACCUMULATE_ATTRIBUTE,
            quantity=quantity,
            attribute=attribute,
            source_player=source_player,
            inverted=inverted,
        )

    def research_technology(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.RESEARCH_TECHNOLOGY,
            source_player=source_player,
            technology=technology,
            inverted=inverted,
        )

    def timer(
            self,
            timer: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.TIMER,
            timer=timer,
            inverted=inverted,
        )

    def object_selected(
            self,
            unit_object: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_SELECTED,
            unit_object=unit_object,
            inverted=inverted,
        )

    def ai_signal(
            self,
            ai_signal: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.AI_SIGNAL,
            ai_signal=ai_signal,
            inverted=inverted,
        )

    def player_defeated(
            self,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.PLAYER_DEFEATED,
            source_player=source_player,
            inverted=inverted,
        )

    def object_has_target(
            self,
            unit_object: Union[int, None] = None,
            next_object: Union[int, None] = None,
            object_list: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_HAS_TARGET,
            unit_object=unit_object,
            next_object=next_object,
            object_list=object_list,
            object_group=object_group,
            object_type=object_type,
            inverted=inverted,
        )

    def object_visible(
            self,
            unit_object: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_VISIBLE,
            unit_object=unit_object,
        )

    def object_not_visible(
            self,
            unit_object: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_NOT_VISIBLE,
            unit_object=unit_object,
        )

    def researching_tech(
            self,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.RESEARCHING_TECH,
            source_player=source_player,
            technology=technology,
            inverted=inverted,
        )

    def units_garrisoned(
            self,
            quantity: Union[int, None] = None,
            unit_object: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.UNITS_GARRISONED,
            quantity=quantity,
            unit_object=unit_object,
            inverted=inverted,
        )

    def difficulty_level(
            self,
            quantity: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.DIFFICULTY_LEVEL,
            quantity=quantity,
            inverted=inverted,
        )

    def chance(
            self,
            quantity: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.CHANCE,
            quantity=quantity,
        )

    def technology_state(
            self,
            quantity: Union[int, None] = None,
            source_player: Union[int, None] = None,
            technology: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.TECHNOLOGY_STATE,
            quantity=quantity,
            source_player=source_player,
            technology=technology,
            inverted=inverted,
        )

    def variable_value(
            self,
            quantity: Union[int, None] = None,
            inverted: Union[int, None] = None,
            variable: Union[int, None] = None,
            comparison: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.VARIABLE_VALUE,
            quantity=quantity,
            inverted=inverted,
            variable=variable,
            comparison=comparison,
        )

    def object_hp(
            self,
            quantity: Union[int, None] = None,
            unit_object: Union[int, None] = None,
            inverted: Union[int, None] = None,
            comparison: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_HP,
            quantity=quantity,
            unit_object=unit_object,
            inverted=inverted,
            comparison=comparison,
        )

    def diplomacy_state(
            self,
            quantity: Union[int, None] = None,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
            target_player: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.DIPLOMACY_STATE,
            quantity=quantity,
            source_player=source_player,
            inverted=inverted,
            target_player=target_player,
        )

    def script_call(
            self,
            xs_function: Union[str, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.SCRIPT_CALL,
            xs_function=xs_function,
        )

    def object_visible_multiplayer(
            self,
            unit_object: Union[int, None] = None,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_VISIBLE_MULTIPLAYER,
            unit_object=unit_object,
            source_player=source_player,
            inverted=inverted,
        )

    def object_selected_multiplayer(
            self,
            unit_object: Union[int, None] = None,
            source_player: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_SELECTED_MULTIPLAYER,
            unit_object=unit_object,
            source_player=source_player,
            inverted=inverted,
        )

    def object_has_action(
            self,
            unit_object: Union[int, None] = None,
            next_object: Union[int, None] = None,
            object_list: Union[int, None] = None,
            object_group: Union[int, None] = None,
            object_type: Union[int, None] = None,
            inverted: Union[int, None] = None,
            unit_ai_action: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OBJECT_HAS_ACTION,
            unit_object=unit_object,
            next_object=next_object,
            object_list=object_list,
            object_group=object_group,
            object_type=object_type,
            inverted=inverted,
            unit_ai_action=unit_ai_action,
        )

    def or_(
            self,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.OR,
        )

    def ai_signal_multiplayer(
            self,
            ai_signal: Union[int, None] = None,
            inverted: Union[int, None] = None,
    ) -> Condition:
        return self._trigger_ref._add_condition(
            ConditionId.AI_SIGNAL_MULTIPLAYER,
            ai_signal=ai_signal,
            inverted=inverted,
        )



from __future__ import annotations

from enum import IntEnum
from typing import List


class TerrainId(IntEnum):
    """
    This enum class provides information about most of the terrains in the game. Information about the
    following properties of a terrain is found in this class:
     - Terrain ID

    **Methods**

    >>> TerrainId.beach_terrains()
    >>> TerrainId.tree_terrains()
    >>> TerrainId.water_terrains()

    **Examples**

    >>> TerrainId.FARM
    >>> 7
    """

    @staticmethod
    def water_terrains() -> List[TerrainId]:
        """
        Returns:
            A list of all Water terrains
        """
        return [
            TerrainId.WATER_DEEP,
            TerrainId.WATER_AZURE,
            TerrainId.WATER_BROWN,
            TerrainId.WATER_GREEN,
            TerrainId.WATER_MEDIUM,
            TerrainId.WATER_2D_BRIDGE,
            TerrainId.WATER_2D_SHORELESS,
            TerrainId.WATER_DEEP_OCEAN,
            TerrainId.WATER_SHALLOW,
        ]

    @staticmethod
    def beach_terrains() -> List[TerrainId]:
        """
        Returns:
            A list of all Beach terrains
        """
        return [
            TerrainId.BEACH,
            TerrainId.BEACH_ICE,
            TerrainId.BEACH_VEGETATION,
            TerrainId.BEACH_WET,
            TerrainId.BEACH_WET_GRAVEL,
            TerrainId.BEACH_WET_ROCK,
            TerrainId.BEACH_WHITE,
            TerrainId.BEACH_WHITE_VEGETATION,
        ]

    @staticmethod
    def tree_terrains() -> List[TerrainId]:
        """
        Returns:
            A list of all Tree terrains
        """
        return [
            TerrainId.FOREST_ACACIA,
            TerrainId.FOREST_AUTUMN,
            TerrainId.FOREST_AUTUMN_SNOW,
            TerrainId.FOREST_BAMBOO,
            TerrainId.FOREST_BAOBAB,
            TerrainId.FOREST_BIRCH,
            TerrainId.FOREST_BUSH,
            TerrainId.FOREST_DEAD,
            TerrainId.FOREST_DRAGON_TREE,
            TerrainId.FOREST_JUNGLE,
            TerrainId.FOREST_MANGROVE,
            TerrainId.FOREST_MEDITERRANEAN,
            TerrainId.FOREST_OAK,
            TerrainId.FOREST_OAK_BUSH,
            TerrainId.FOREST_PALM_DESERT,
            TerrainId.FOREST_PINE,
            TerrainId.FOREST_PINE_SNOW,
            TerrainId.FOREST_RAINFOREST,
            TerrainId.FOREST_REEDS,
            TerrainId.FOREST_REEDS_BEACH,
            TerrainId.FOREST_REEDS_SHALLOWS,
        ]

    BEACH = 2
    BEACH_NON_NAVIGABLE = 79
    BEACH_NON_NAVIGABLE_WET_GRAVEL = 81
    BEACH_NON_NAVIGABLE_WET_ROCK = 82
    BEACH_NON_NAVIGABLE_WET_SAND = 80
    BEACH_ICE = 37
    BEACH_VEGETATION = 52
    BEACH_WET = 107
    BEACH_WET_GRAVEL = 108
    BEACH_WET_ROCK = 109
    BEACH_WHITE = 53
    BEACH_WHITE_VEGETATION = 51
    BLACK = 47
    DESERT_CRACKED = 45
    DESERT_QUICKSAND = 46
    DESERT_SAND = 14
    DIRT_1 = 6
    DIRT_2 = 11
    DIRT_3 = 3
    DIRT_4 = 42
    DIRT_MUD = 76
    DIRT_SAVANNAH = 41
    FARM = 7
    FARM_0 = 29
    FARM_33 = 30
    FARM_67 = 31
    FARM_DEAD = 8
    FOREST_ACACIA = 50
    FOREST_AUTUMN = 104
    FOREST_AUTUMN_SNOW = 105
    FOREST_BAMBOO = 18
    FOREST_BAOBAB = 49
    FOREST_BIRCH = 110
    FOREST_BUSH = 89
    FOREST_DEAD = 106
    FOREST_DRAGON_TREE = 48
    FOREST_JUNGLE = 17
    FOREST_MANGROVE = 55
    FOREST_MEDITERRANEAN = 88
    FOREST_OAK = 10
    FOREST_OAK_BUSH = 20
    FOREST_PALM_DESERT = 13
    FOREST_PINE = 19
    FOREST_PINE_SNOW = 21
    FOREST_RAINFOREST = 56
    FOREST_REEDS = 92
    FOREST_REEDS_BEACH = 91
    FOREST_REEDS_SHALLOWS = 90
    GRASS_1 = 0
    GRASS_2 = 12
    GRASS_3 = 9
    GRASS_DRY = 100
    GRASS_FOUNDATION = 27
    GRASS_JUNGLE = 60
    GRASS_JUNGLE_RAINFOREST = 83
    GRASS_OTHER = 16
    GRAVEL_DEFAULT = 70
    GRAVEL_DESERT = 102
    ICE = 35
    ICE_NAVIGABLE = 26
    MODDABLE_GRASS_1 = 84
    MODDABLE_GRASS_2 = 85
    MODDABLE_GRASS_3 = 86
    MODDABLE_GRASS_4 = 87
    MODDABLE_DEEP_WATER = 99
    MODDABLE_NORMAL_WATER_1 = 97
    MODDABLE_NORMAL_WATER_2 = 98
    MODDABLE_SHALLOWS_1 = 93
    MODDABLE_SHALLOWS_2 = 94
    OBSOLETE_LEAVES_JUNGLE = 62
    OBSOLETE_ROAD_DESERT = 43
    OBSOLETE_ROAD_FUNGUS = 39
    OBSOLETE_ROAD_GRAVEL = 103
    OBSOLETE_ROAD_JUNGLE = 61
    OBSOLETE_ROAD_SNOW = 38
    OBSOLETE_SNOW_DIRT = 33
    OBSOLETE_SNOW_GRASS = 34
    RESERVED = 68
    RICE_FARM = 63
    RICE_FARM_0 = 65
    RICE_FARM_33 = 66
    RICE_FARM_67 = 67
    RICE_FARM_DEAD = 64
    ROAD = 24
    ROAD_BROKEN = 25
    ROAD_FUNGUS = 75
    ROAD_GRAVEL = 78
    ROCK_1 = 40
    SHALLOWS = 4
    SHALLOWS_AZURE = 59
    SHALLOWS_MANGROVE = 54
    SNOW = 32
    SNOW_FOUNDATION = 36
    SNOW_LIGHT = 73
    SNOW_STRONG = 74
    SWAMP = 111
    SWAMP_BOGLAND = 101
    UNDERBUSH = 5
    UNDERBUSH_JUNGLE = 77
    UNDERBUSH_LEAVES = 71
    UNDERBUSH_SNOW = 72
    VERY_EVIL_FOG = 69
    WATER_2D_BRIDGE = 28
    WATER_2D_SHORELESS = 15
    WATER_AZURE = 58
    WATER_BROWN = 96
    WATER_DEEP = 22
    WATER_DEEP_OCEAN = 57
    WATER_GREEN = 95
    WATER_MEDIUM = 23
    WATER_SHALLOW = 1



from typing import List

from AoE2ScenarioParser.helper.exceptions import InvalidScenarioStructureError
from AoE2ScenarioParser.helper.pretty_format import pretty_format_list
from AoE2ScenarioParser.helper.string_manipulations import add_tabs


class DependencyTarget:
    def __init__(self, targets: List[List[str]]):
        """
        Object for targeting a specific retriever based on it's section name (or 'self') and the retriever name.

        Args:
            targets: a list of lists referencing a target according to: "('self' or {section}):{attribute_path}"
        """
        self.targets = targets

    @classmethod
    def instance_or_none(cls, target):
        if target is None:
            return None

        if type(target) is str:
            return cls([target.split(":")])
        elif type(target) is list:
            return cls([entry.split(':') for entry in target])
        else:
            raise InvalidScenarioStructureError(
                "Target defined using unknown type. For single targets, use str else list"
            )

    def __repr__(self) -> str:
        return f"[DependencyTarget] {add_tabs(pretty_format_list(self.targets).strip(), 1)}"



import sys

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.printers import warn

_block_below_minor = 8
_notify_below_minor = 8
_py_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"


def python_version_check():
    if sys.version_info[0] < 3:
        raise Exception("AoE2ScenarioParser requires the use of Python 3")

    if sys.version_info[1] < _block_below_minor:
        raise Exception(f"\n\nAoE2ScenarioParser requires the use of Python 3.{_block_below_minor}.\n"
                        f"You are currently using version {_py_version}. You can download a newer version from: \n"
                        f"\thttps://www.python.org/downloads/\n")
    elif sys.version_info[1] < _notify_below_minor:
        if not settings.DISABLE_VERSION_WARNINGS:
            warn(f"AoE2ScenarioParser will update it's minimum python requirements to Python 3.{_notify_below_minor} soon.\n"
                 f"You are currently using version {_py_version}. You can download a newer version from: \n"
                 f"\thttps://www.python.org/downloads/\n\n"
                 f"You can disable these warnings using:\n" 
                 "    from AoE2ScenarioParser import settings\n" 
                 "    settings.DISABLE_VERSION_WARNINGS = True")



import math
from uuid import UUID

from AoE2ScenarioParser.scenarios.scenario_store import getters

from AoE2ScenarioParser.sections.dependencies.dependency_action import DependencyAction
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever


def refresh_targets(retriever_event, section, host_uuid):
    for target in retriever_event.dependency_target.targets:
        selected_retriever = select_retriever(target, section, host_uuid)
        # selected_retriever = get_retriever_by_name(retriever_list, target[1])
        # selected_retriever = section.retriever_map[target[1]]
        execute_refresh_action(selected_retriever, section, host_uuid)


def execute_refresh_action(retriever, section, host_uuid):
    handle_retriever_dependency(retriever, "refresh", section, host_uuid)


def handle_retriever_dependency(retriever: Retriever, state, section, host_uuid: UUID):
    on_x = f'on_{state}'
    if not hasattr(retriever, on_x):
        return

    retriever_event = getattr(retriever, on_x)  # construct, commit or refresh

    action = retriever_event.dependency_action

    if action == DependencyAction.REFRESH_SELF:
        execute_refresh_action(retriever, section, host_uuid)
    elif action == DependencyAction.REFRESH:
        refresh_targets(retriever_event, section, host_uuid)
    elif action in [DependencyAction.SET_VALUE, DependencyAction.SET_REPEAT]:
        value = execute_dependency_eval(retriever_event, section, host_uuid)
        if action == DependencyAction.SET_VALUE:
            retriever.data = value
        elif action == DependencyAction.SET_REPEAT:
            retriever.datatype.repeat = value


def execute_dependency_eval(retriever_event, section, host_uuid):
    eval_code = retriever_event.dependency_eval.eval_code
    eval_locals = retriever_event.dependency_eval.eval_locals
    targets = retriever_event.dependency_target.targets

    values = []
    for target in targets:
        # retriever_list = select_retriever_list(target, self_list, sections)
        # values.append(get_retriever_by_name(retriever_list, target[1]).data)
        values.append(select_retriever(target, section, host_uuid).data)

    for index, target in enumerate(targets):
        eval_locals[target[1]] = values[index]
    eval_locals['math'] = math

    return eval(eval_code, {}, eval_locals)


def select_retriever(target, section, host_uuid):
    if target[0] == "self":
        return section.retriever_map[target[1]]
    else:
        sections = getters.get_sections(host_uuid)
        return sections[target[0]].retriever_map[target[1]]



from AoE2ScenarioParser.objects.data_objects.trigger import Trigger


class TriggerSelect:
    def __init__(self, trigger_index=None, display_index=None, trigger=None):
        """
        Object used to select a trigger in many trigger related functions. For ease of use, the alias `TS` can be
        called. You can also use those in combination with the class methods (factory methods). Like so:

        TS.index(4)    # To select the trigger with index 4

        TS.display(4)  # Trigger with display index 4

        TS.trigger(trigger)  # Well... The trigger object given...

        Args:
            trigger_index (int): The index of the trigger. Starting from 0, based on creation time
            display_index (int): The display index of a trigger. Starting from 0, based on display order in the editor
            trigger (Trigger): The trigger object itself.
        """
        self.trigger = trigger
        self.display_index = display_index
        self.trigger_index = trigger_index

    @classmethod
    def index(cls, index: int):
        return cls(trigger_index=index)

    @classmethod
    def display(cls, display_index: int):
        return cls(display_index=display_index)

    @classmethod
    def trigger(cls, trigger: Trigger):
        return cls(trigger=trigger)


TS = TriggerSelect



from typing import Optional, List, Tuple, Dict, TYPE_CHECKING, Union
from uuid import UUID

from AoE2ScenarioParser.scenarios.scenario_store import store

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.data_objects.unit import Unit
    from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
    from AoE2ScenarioParser.scenarios.aoe2_de_scenario import AoE2DEScenario
    from AoE2ScenarioParser.objects.data_objects.terrain_tile import TerrainTile
    from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection


def get_unit(uuid: UUID, unit_reference_id: int) -> Optional['Unit']:
    """
    Get a placed unit based on it's reference id in a scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        unit_reference_id (int): The reference_id of the unit

    Returns:
        The Unit Object
    """
    units, _ = get_units(uuid, [unit_reference_id])
    if units:
        return units[0]
    return None


def get_units_in_area(uuid: UUID, x1: int, y1: int, x2: int, y2: int) -> Optional[List['Unit']]:
    """
    Get a placed unit based on it's reference id in a scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        x1: The X location of the left corner
        y1: The Y location of the left corner
        x2: The X location of the right corner
        y2: The Y location of the right corner

    Returns:
        The Unit Object
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.unit_manager.get_units_in_area(x1, y1, x2, y2)
    return None


def get_units(uuid: UUID, unit_reference_ids: List[int]) -> Optional[Tuple[List['Unit'], List[int]]]:
    """
    Get a placed unit based on it's reference id in a scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        unit_reference_ids (List[int]): The reference_ids of the units

    Returns:
        A tuple with a list of the found unit objects and a list of the IDs that weren't found.
    """
    unit_reference_ids = unit_reference_ids.copy()
    scenario = store.get_scenario(uuid)
    if scenario:
        result = []
        for unit in scenario.unit_manager.get_all_units():
            if unit.reference_id in unit_reference_ids:
                unit_reference_ids.remove(unit.reference_id)
                result.append(unit)
        return result, unit_reference_ids
    return None


def get_sections(uuid: UUID) -> Optional[Dict[str, 'AoE2FileSection']]:
    """
    Get the section dict of a scenario.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The sections of the selected scenario
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.sections
    return None


def get_scenario_version(uuid: UUID) -> Optional[str]:
    """
    Get the scenario version.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The scenario version of the selected scenario (e.g. '1.43')
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.scenario_version
    return None


def get_game_version(uuid: UUID) -> Optional[str]:
    """
    Get the game version.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The game version of the selected scenario (e.g. 'DE')
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.game_version
    return None


def get_map_size(uuid: UUID) -> Optional[int]:
    """
    Get the map size of a scenario. Scenario is selected based on the given UUID.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The map size of the scenario
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.map_manager.map_size
    return None


def get_terrain(uuid: UUID) -> Optional[List['TerrainTile']]:
    """
    Get the map size of a scenario. Scenario is selected based on the given UUID.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The map size of the scenario
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.map_manager.terrain
    return None


def get_trigger(uuid: UUID, trigger_index: int) -> Optional['Trigger']:
    """
    Get a trigger in a scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        trigger_index (int): The index of the trigger

    Returns:
        The trigger with the given ID
    """
    scenario = store.get_scenario(uuid)
    if scenario and trigger_index < len(scenario.trigger_manager.triggers):
        return scenario.trigger_manager.triggers[trigger_index]
    return None


def get_triggers_by_prefix(uuid: UUID, prefix: Union[str, Tuple]) -> Optional[List['Trigger']]:
    """
    Get the trigger version of the scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        prefix: The prefix to check trigger names against

    Returns:
        Triggers matching the given prefix
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return [trigger for trigger in scenario.trigger_manager.triggers if trigger.name.startswith(prefix)]
    return None


def get_variable_name(uuid: UUID, variable_index: int) -> Optional[str]:
    """
    Get the variable name in a scenario.

    Args:
        uuid (UUID): The UUID of the scenario
        variable_index (int): The index of the variable

    Returns:
        The name of the variable with the given ID
    """
    scenario: Optional[AoE2DEScenario] = store.get_scenario(uuid)
    if scenario:
        if gv := get_game_version(uuid) == "DE":
            variable = scenario.trigger_manager.get_variable(variable_index)
            if variable:
                return variable.name
            elif 0 <= variable_index <= 255:
                return f"Variable {variable_index}"
        else:
            raise ValueError(f"Scenarios with the game version: {gv} do not support variables.")
    return None


def get_trigger_version(uuid: UUID) -> Optional[float]:
    """
    Get the trigger version of the scenario.

    Args:
        uuid (UUID): The UUID of the scenario

    Returns:
        The trigger version.
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.sections['Triggers'].trigger_version
    return None



from typing import List

from AoE2ScenarioParser.helper.helper import exclusive_if
from AoE2ScenarioParser.objects.data_objects.trigger import Trigger
from AoE2ScenarioParser.objects.data_objects.variable import Variable
from AoE2ScenarioParser.objects.managers.trigger_manager import TriggerManager
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class TriggerManagerDE(TriggerManager):
    _link_list = [
        RetrieverObjectLink("triggers", "Triggers", "trigger_data", process_as_object=Trigger),
        RetrieverObjectLink("trigger_display_order", "Triggers", "trigger_display_order_array"),
        RetrieverObjectLink("variables", "Triggers", "variable_data", process_as_object=Variable),
    ]

    def __init__(self, triggers: List[Trigger], trigger_display_order: List[int], variables: List[Variable], **kwargs):
        self.variables: List[Variable] = variables

        super().__init__(triggers, trigger_display_order, **kwargs)

    def add_variable(self, name: str, variable_id: int = -1) -> Variable:
        """
        Adds a variable.

        Args:
            name (str): The name for the variable
            variable_id (int): The ID of the variable. If left empty (default: -1), lowest available value will be used

        Returns:
            The newly added Variable
        """
        list_of_var_ids = [var.variable_id for var in self.variables]
        if variable_id == -1:
            for i in range(256):
                if i not in list_of_var_ids:
                    variable_id = i
                    break
            if variable_id == -1:
                raise IndexError(f"No variable ID available. All in use? In use: ({list_of_var_ids}/256)")
        if not (0 <= variable_id <= 255):
            raise ValueError("Variable ID has to fall between 0 and 255 (incl).")
        if variable_id in list_of_var_ids:
            raise ValueError("Variable ID already in use.")

        new_variable = Variable(variable_id=variable_id, name=name, host_uuid=self._host_uuid)
        self.variables.append(new_variable)
        return new_variable

    def get_variable(self, variable_id: int = None, variable_name: str = None) -> Variable:
        if not exclusive_if(variable_id is not None, variable_name is not None):
            raise ValueError("Select a variable using either the variable_id or variable_name parameters.")
        for variable in self.variables:
            if variable.variable_id == variable_id or variable.name == variable_name:
                return variable

    def get_summary_as_string(self) -> str:
        return_string = super().get_summary_as_string()

        return_string += "\nVariables Summary:\n"
        if len(self.variables) == 0:
            return_string += "\t<< No Variables >>"

        longest_variable_name = -1
        for variable in self.variables:
            longest_variable_name = max(longest_variable_name, len(variable.name))

        longest_variable_name += 3
        for index, variable in enumerate(self.variables):
            var_name = variable.name
            name_buffer = " " * (longest_variable_name - len(var_name))
            return_string += f"\t{var_name}{name_buffer}[Index: {variable.variable_id}]\n"

        return return_string

    def get_content_as_string(self) -> str:
        return_string = super().get_content_as_string()

        return_string += "Variables:\n"

        if len(self.variables) == 0:
            return_string += "\t<<No Variables>>\n"

        for variable in self.variables:
            return_string += f"\t'{variable.name}' [Index: {variable.variable_id}]\n"

        return return_string



from __future__ import annotations

import itertools
from typing import List, Union, Tuple, Set, Optional

from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.helper import xy_to_i
from AoE2ScenarioParser.helper.list_functions import list_chuncks
from AoE2ScenarioParser.helper.maffs import sign
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.data_objects.terrain_tile import TerrainTile, reset_terrain_index
from AoE2ScenarioParser.objects.support.uuid_list import UuidList
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink


class MapManager(AoE2Object):
    """Manager of the everything map related."""

    _link_list = [
        RetrieverObjectLink("map_width", "Map", "map_width"),
        RetrieverObjectLink("map_height", "Map", "map_height"),
        RetrieverObjectLink("terrain", "Map", "terrain_data", process_as_object=TerrainTile),
    ]

    def __init__(self,
                 map_width: int,
                 map_height: int,
                 terrain: List[TerrainTile],
                 **kwargs
                 ):
        super().__init__(**kwargs)

        self._map_width = map_width
        self._map_height = map_height
        self.terrain = terrain

    @property
    def terrain_2d(self) -> List[List[TerrainTile]]:
        return list(list_chuncks(self.terrain, self.map_size))

    def get_tile(self, x: int = None, y: int = None, i: int = None) -> TerrainTile:
        if i and (x or y):
            raise ValueError("Cannot use both xy and i. Choose or XY or I.")
        if i is not None:
            return self.terrain[i]
        return self.terrain[xy_to_i(x, y, self.map_size)]

    def get_tile_safe(self, x: int = None, y: int = None, i: int = None) -> Union[TerrainTile, None]:
        if i and (x or y):
            raise ValueError("Cannot use both xy and i. Choose or XY or I.")
        try:
            return self.get_tile(x, y, i)
        except (IndexError, ValueError):
            return None

    def get_square_1d(self, x1, y1, x2, y2) -> List[TerrainTile]:
        """
        Get a square of tiles from the map

        Args:
            x1 (int): The x1 coordinate of the square
            y1 (int): The y1 coordinate of the square
            x2 (int): The x2 coordinate of the square
            y2 (int): The y2 coordinate of the square

        Returns:
            1D list of terrain tiles based on given coordinates
        """
        result = []
        for row in self._get_square_rows(x1, y1, x2, y2):
            result.extend(row)
        return result

    def get_square_2d(self, x1, y1, x2, y2) -> List[List[TerrainTile]]:
        """
        Get a square of tiles from the map

        Args:
            x1 (int): The x1 coordinate of the square
            y1 (int): The y1 coordinate of the square
            x2 (int): The x2 coordinate of the square
            y2 (int): The y2 coordinate of the square

        Returns:
            2D list of lists with terrain tiles based on given coordinates
        """
        result = []
        for row in self._get_square_rows(x1, y1, x2, y2):
            result.append(row)
        return result

    def _get_square_rows(self, x1, y1, x2, y2):
        row_nums = range(y1, y2 + 1)
        for row in row_nums:
            i1 = xy_to_i(x1, row, self.map_size)
            i2 = xy_to_i(x2, row, self.map_size)
            tiles = self.terrain[i1:i2 + 1]
            yield tiles

    @property
    def map_width(self) -> int:
        return self._map_width

    @property
    def map_height(self) -> int:
        return self._map_height

    @property
    def map_size(self) -> int:
        if self._map_height == self._map_width:
            return self._map_height
        else:
            raise ValueError("Map is not a square. Use the attributes 'map_width' and 'map_height' instead.")

    @property
    def terrain(self) -> List[TerrainTile]:
        return self._terrain

    @terrain.setter
    def terrain(self, value: List[TerrainTile]):
        def reset_indices(lst):
            tile: TerrainTile
            for index, tile in enumerate(lst):
                reset_terrain_index(tile, index)

        if value is not None:
            self._terrain = UuidList(
                uuid=self._host_uuid,
                seq=value,
                on_update_execute_list=reset_indices
            )

    @map_size.setter
    def map_size(self, new_size: int):
        old_size = self._map_width
        difference = new_size - old_size

        self._map_width = new_size
        self._map_height = new_size

        new_terrain = []
        if difference < 0:
            # Remove ends of rows (x) & remove final rows entirely (y)
            for index, chunck in enumerate(list_chuncks(self.terrain, old_size)):
                if index == new_size:
                    break
                new_terrain.extend(chunck[:new_size])
        elif difference > 0:
            # Add ends to rows (x) & add entirely new rows  (y)
            chunck_gen = list_chuncks(self.terrain, old_size)
            for index in range(new_size):
                if index < old_size:
                    row = next(chunck_gen) + [TerrainTile(host_uuid=self._host_uuid) for _ in range(difference)]
                else:
                    row = [TerrainTile(host_uuid=self._host_uuid) for _ in range(new_size)]
                new_terrain.extend(row)
        self.terrain = new_terrain

    def set_elevation(self, elevation: int, x1: int, y1: int, x2: Optional[int] = None,
                      y2: Optional[int] = None) -> None:
        """
        Sets elevation like the in-game elevation mechanics. Can set a hill (or single point) to a certain height and
        all tiles around it are adjusted accordingly.

        If you find that the in-game mechanics work differently than this function please report it.

        Args:
            elevation (int): The elevation to create at the coordinates
            x1 (int): The x coordinate of the west corner
            y1 (int): The y coordinate of the west corner
            x2 (Optional[int]): The x coordinate of the east corner
            y2 (Optional[int]): The y coordinate of the east corner
        """
        x2 = x1 if x2 is None else x2
        y2 = y1 if y2 is None else y2

        if x1 == x2 and y1 == y2:
            edge_tiles = source_tiles = [self.get_tile(x1, y1)]
            xys = [source_tiles[0].xy]
        else:
            source_tiles = self.get_square_2d(x1, y1, x2, y2)
            xys = set(t.xy for row in source_tiles for t in row)
            edge_tiles = source_tiles[0] + source_tiles[-1]
            for i in range(1, len(source_tiles) - 1):
                edge_tiles.extend([source_tiles[i][0], source_tiles[i][-1]])

        for tile in edge_tiles:
            tile.elevation = elevation
            self._elevation_tile_recursion(tile, xys)

    def _elevation_tile_recursion(
            self,
            source_tile: TerrainTile,
            xys: Set[Tuple[int, int]],
            visited: Set[Tuple[int, int]] = None
    ):
        """
        Elevation recursive function. Used in the set_elevation function

        Args:
            source_tile (TerrainTile): The tile to check around
            xys (Set[Tuple[int, int]]): The XY tuples from the initial square
            visited (Set[Tuple[int, int]]): The visited XY tuples with this recursion tree path
        """
        visited = set() if visited is None else visited.copy()
        x, y = source_tile.xy
        visited.add((x, y))
        for nx, ny in itertools.product(range(-1, 2), repeat=2):
            new_x, new_y = x + nx, y + ny
            if (nx or ny) and (new_x, new_y) not in xys and (new_x, new_y) not in visited:
                other = self.get_tile_safe(new_x, new_y)
                if other is None:
                    continue
                behind = self.get_tile_safe(x + nx * 2, y + ny * 2)
                if behind is not None and other.elevation < source_tile.elevation == behind.elevation:
                    other.elevation = source_tile.elevation
                elif abs(other.elevation - source_tile.elevation) > 1:
                    other.elevation = source_tile.elevation + int(sign(other.elevation, source_tile.elevation))
                    self._elevation_tile_recursion(other, xys, visited)

    def create_hill(self, x1, y1, x2, y2, elevation) -> None:
        """
        Function that takes the coordinates and the height of a plateau and applies it to the map
        by also setting the surrounding slopes so that it is smooth.

        **This function can only increase height. It will not lower areas terrain.
        For that, you can use: `map_manager.set_elevation()`.**

        Args:
            x1 (int): The x coordinate of the west corner
            y1 (int): The y coordinate of the west corner
            x2 (int): The x coordinate of the east corner
            y2 (int): The y coordinate of the east corner
            elevation (int): The elevation of the map. Default in-game = 0 (called 1 in the game), in-game max = 6
                (called 7 in game). If the given value is over 20 the game camera will 'clip' into the hill.
                So the in-game camera hovers around the height of 20/21 when fully zoomed in, without Ultra Graphics.

        :Author:
            pvallet
        """
        warn(f"The function `MapManager.create_hill()` is deprecated as of 0.1.27. "
             f"It will be removed in the future. Please use map_manager.set_elevation() instead.")

        for x in range(max(0, x1 - elevation), min(self.map_size, x2 + elevation)):
            for y in range(max(0, y1 - elevation), min(self.map_size, y2 + elevation)):
                if x1 <= x <= x2 and y1 <= y <= y2:
                    intended_elevation = elevation
                else:
                    distance_to_hill = max(x1 - x, x - x2, y1 - y, y - y2)
                    intended_elevation = elevation - distance_to_hill

                tile = self.terrain[helper.xy_to_i(x, y, self.map_size)]
                tile.elevation = max(intended_elevation, tile.elevation)



from __future__ import annotations

from enum import IntEnum
from typing import Union

from AoE2ScenarioParser.datasets import conditions
from AoE2ScenarioParser.helper.helper import raise_if_not_int_subclass, validate_coords, value_is_valid
from AoE2ScenarioParser.helper.printers import warn
from AoE2ScenarioParser.helper.string_manipulations import add_tabs
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.support.attr_presentation import transform_condition_attr_value
from AoE2ScenarioParser.sections.retrievers.retriever_object_link import RetrieverObjectLink
from AoE2ScenarioParser.sections.retrievers.support import Support


class Condition(AoE2Object):
    """Object for handling a condition."""

    _link_list = [
        RetrieverObjectLink("condition_type", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].condition_type"),
        RetrieverObjectLink("quantity", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].quantity"),
        RetrieverObjectLink("attribute", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].attribute"),
        RetrieverObjectLink("unit_object", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].unit_object"),
        RetrieverObjectLink("next_object", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].next_object"),
        RetrieverObjectLink("object_list", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].object_list"),
        RetrieverObjectLink("source_player", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].source_player"),
        RetrieverObjectLink("technology", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].technology"),
        RetrieverObjectLink("timer", "Triggers", "trigger_data[__index__].condition_data[__index__].timer"),
        RetrieverObjectLink("area_x1", "Triggers", "trigger_data[__index__].condition_data[__index__].area_x1"),
        RetrieverObjectLink("area_y1", "Triggers", "trigger_data[__index__].condition_data[__index__].area_y1"),
        RetrieverObjectLink("area_x2", "Triggers", "trigger_data[__index__].condition_data[__index__].area_x2"),
        RetrieverObjectLink("area_y2", "Triggers", "trigger_data[__index__].condition_data[__index__].area_y2"),
        RetrieverObjectLink("object_group", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].object_group"),
        RetrieverObjectLink("object_type", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].object_type"),
        RetrieverObjectLink("ai_signal", "Triggers", "trigger_data[__index__].condition_data[__index__].ai_signal"),
        RetrieverObjectLink("inverted", "Triggers", "trigger_data[__index__].condition_data[__index__].inverted"),
        RetrieverObjectLink("variable", "Triggers", "trigger_data[__index__].condition_data[__index__].variable"),
        RetrieverObjectLink("comparison", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].comparison"),
        RetrieverObjectLink("target_player", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].target_player"),
        RetrieverObjectLink("object_state", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].object_state", Support(since=1.42)),
        RetrieverObjectLink("xs_function", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].xs_function", Support(since=1.40)),
        RetrieverObjectLink("unit_ai_action", "Triggers",
                            "trigger_data[__index__].condition_data[__index__].unit_ai_action", Support(since=1.40)),
    ]

    def __init__(self,
                 condition_type: int = None,
                 quantity: int = None,
                 attribute: int = None,
                 unit_object: int = None,
                 next_object: int = None,
                 object_list: int = None,
                 source_player: Union[int, IntEnum] = None,
                 technology: Union[int, IntEnum] = None,
                 timer: int = None,
                 area_x1: int = None,
                 area_y1: int = None,
                 area_x2: int = None,
                 area_y2: int = None,
                 object_group: int = None,
                 object_type: int = None,
                 ai_signal: int = None,
                 inverted: int = None,
                 variable: int = None,
                 comparison: int = None,
                 target_player: Union[int, IntEnum] = None,
                 unit_ai_action: int = None,
                 object_state: int = None,
                 xs_function: str = None,
                 **kwargs
                 ):
        raise_if_not_int_subclass([object_list, technology])
        area_x1, area_y1, area_x2, area_y2 = validate_coords(area_x1, area_y1, area_x2, area_y2)

        self.condition_type: int = condition_type
        self.quantity: int = quantity
        self.attribute: int = attribute
        self.unit_object: int = unit_object
        self.next_object: int = next_object
        self.object_list: int = object_list
        self.source_player: int = source_player
        self.technology: int = technology
        self.timer: int = timer
        self.area_x1: int = area_x1
        self.area_y1: int = area_y1
        self.area_x2: int = area_x2
        self.area_y2: int = area_y2
        self.object_group: int = object_group
        self.object_type: int = object_type
        self.ai_signal: int = ai_signal
        self.inverted: int = inverted
        self.variable: int = variable
        self.comparison: int = comparison
        self.target_player: int = target_player
        self.unit_ai_action: int = unit_ai_action
        self.object_state: int = object_state
        self.xs_function: str = xs_function

        super().__init__(**kwargs)

    def get_content_as_string(self, include_effect_definition=False) -> str:
        if self.condition_type not in conditions.attributes:
            attributes_list = conditions.empty_attributes
        else:
            attributes_list = conditions.attributes[self.condition_type]

        return_string = ""
        for attribute in attributes_list:
            val = getattr(self, attribute)
            if attribute == "condition_type" or val in [[], [-1], [''], "", " ", -1]:
                continue

            value_string = transform_condition_attr_value(self.condition_type, attribute, val, self._host_uuid)
            return_string += f"{attribute}: {value_string}\n"

        if return_string == "":
            return "<< No Attributes >>\n"

        if include_effect_definition:
            return f"{conditions.condition_names[self.condition_type]}:\n{add_tabs(return_string, 1)}"

        return return_string

    def __str__(self):
        return f"[Condition] {self.get_content_as_string(include_effect_definition=True)}"



from typing import Type, List

from AoE2ScenarioParser import settings
from AoE2ScenarioParser.helper.exceptions import UnsupportedAttributeError
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.scenarios.scenario_store import getters
from AoE2ScenarioParser.sections.aoe2_file_section import AoE2FileSection
from AoE2ScenarioParser.sections.dependencies.dependency import handle_retriever_dependency
from AoE2ScenarioParser.sections.retrievers.support import Support


class RetrieverObjectLink:
    def __init__(self,
                 variable_name: str,
                 section_name: str = None,
                 link: str = None,
                 support: Support = None,
                 process_as_object: Type[AoE2Object] = None,
                 retrieve_instance_number: bool = False,
                 retrieve_history_number: int = -1,
                 commit_callback=None,
                 destination_object=None,
                 ):
        if sum([link is not None, retrieve_instance_number, (retrieve_history_number != -1)]) != 1:
            raise ValueError("You must use exactly one of 'link' and the two 'retrieve...number' parameters.")
        self.destination_object = destination_object

        self.name: str = variable_name
        self.section_name = section_name
        self.link = link
        self.support: Support = support
        self.is_special_unit_case = self._is_special_unit_case()
        self.process_as_object: Type[AoE2Object] = process_as_object
        self.retrieve_instance_number: bool = retrieve_instance_number
        self.retrieve_history_number: int = retrieve_history_number
        self.commit_callback = commit_callback

        self.splitted_link: List[str] = link.split('.') if link is not None else []

    def construct(self, host_uuid, number_hist=None):
        if number_hist is None:
            number_hist = []
        instance_number = AoE2Object.get_instance_number(number_hist=number_hist)

        if self.retrieve_instance_number:
            return instance_number
        elif self.retrieve_history_number != -1:
            return number_hist[self.retrieve_history_number]
        else:
            if self.is_special_unit_case:
                return self._construct_special_unit_case(host_uuid)

            sections = getters.get_sections(host_uuid)
            scenario_version = getters.get_scenario_version(host_uuid)

            # Retrieve value without using eval() -- Eval is slow
            value = sections[self.section_name]
            for index, item in enumerate(self.splitted_link):
                try:
                    if "[" in item:
                        value = getattr(value, item[:-11])[number_hist[index]]
                    else:
                        value = getattr(value, item)
                except AttributeError as e:
                    # Maybe not supported in current version. if actually not supported -> ignore
                    if self.support is not None:
                        if not self.support.supports(scenario_version):
                            value = None
                            break
                    raise e

            if self.process_as_object:
                return self.process_object_list(value, number_hist, host_uuid)
            return value

    def process_object_list(self, value_list, instance_number_history, host_uuid):
        object_list = []
        for index, struct in enumerate(value_list):
            object_list.append(
                self.process_as_object._construct(host_uuid, instance_number_history + [index])
            )
        return object_list

    def commit(self, host_uuid, host_obj: AoE2Object):
        # Object-only attributes for the ease of access of information.
        # Not actually representing a value in the scenario file.
        if self.retrieve_instance_number or self.retrieve_history_number >= 0:
            return

        number_hist = host_obj._instance_number_history

        try:
            # Get new value for receiver
            value = getattr(host_obj, self.name)
        except UnsupportedAttributeError:
            return  # Not supported in current version.

        if self.commit_callback is not None:
            value = self.commit_callback(host_obj, self.name, value)

        sections = getters.get_sections(host_uuid)
        section = sections[self.section_name]

        if self.is_special_unit_case:
            self._commit_special_unit_case(host_uuid, value)
            return

        # Retrieve value without using eval() -- Eval is slow
        retriever = None
        file_section = section
        for index, item in enumerate(self.splitted_link):
            try:
                if "[" in item:
                    file_section = getattr(file_section, item[:-11])[number_hist[index]]
                else:
                    retriever = file_section.retriever_map[item]
            except KeyError as e:
                # Maybe not supported in current version. if actually not supported -> ignore
                if self.support is not None:
                    if not self.support.supports(
                            getters.get_scenario_version(host_uuid)):
                        return
                if settings.IGNORE_WRITING_ERRORS:
                    return
                raise e

        if retriever is None:
            raise ValueError("RetrieverObjectLink is unable to find retriever")

        if self.process_as_object:
            struct_datatype = retriever.datatype.var

            prefix = "struct:"
            if not struct_datatype.startswith(prefix):
                raise ValueError(
                    f"process_as_object isn't defined properly. Expected: '{prefix}...', got: '{struct_datatype}'"
                )

            struct_name = struct_datatype[len(prefix):]
            struct_model = file_section.struct_models[struct_name]

            RetrieverObjectLink.update_retriever_length(retriever, struct_model, len(value), host_uuid)
            RetrieverObjectLink.commit_object_list(value, host_obj._instance_number_history)
        else:
            retriever.data = value

        if hasattr(retriever, 'on_commit'):
            handle_retriever_dependency(retriever, "commit", file_section, host_uuid)

    @staticmethod
    def commit_object_list(object_list, instance_number_history):
        obj: AoE2Object
        for index, obj in enumerate(object_list):
            obj._instance_number_history = instance_number_history + [index]
            obj.commit()

    @staticmethod
    def update_retriever_length(retriever, model, new_len, host_uuid):
        try:
            old_len = len(retriever.data)
        except TypeError:  # retriever.data was not set before (list of 0 -> None)
            old_len = 0
            retriever.data = []

        if new_len < old_len:
            retriever.data = retriever.data[:new_len]
        elif new_len > old_len:
            retriever.data += [
                AoE2FileSection.from_model(model, host_uuid, set_defaults=True)
                for _ in range(new_len - old_len)
            ]

            if retriever.log_value:
                retriever._print_value_update(f"[{model.name}] * {old_len}", f"[{model.name}] * {new_len}")

    def _is_special_unit_case(self) -> bool:
        return ("[]" in self.link) if self.link else False

    def _construct_special_unit_case(self, host_uuid):
        units = []
        sections = getters.get_sections(host_uuid)
        value = sections[self.section_name]
        for index, item in enumerate(self.splitted_link):
            if "[]" in item:
                value = getattr(value, item[:-2])
            else:
                for player, player_units_section in enumerate(value):
                    player_units = getattr(player_units_section, item)
                    units.append(self.process_object_list(player_units, [player], host_uuid))
        return units

    def _commit_special_unit_case(self, host_uuid, value):
        sections = getters.get_sections(host_uuid)

        for player, player_unit in enumerate(value):
            player_unit_retriever = sections["Units"].players_units[player]
            retriever_list = player_unit_retriever.retriever_map.values()
            units = player_unit_retriever.retriever_map["units"]
            # units = get_retriever_by_name(retriever_list, "units")
            struct_model = player_unit_retriever.struct_models["UnitStruct"]

            RetrieverObjectLink.update_retriever_length(units, struct_model, len(value[player]), host_uuid)
            RetrieverObjectLink.commit_object_list(player_unit, [player])

            for retriever in retriever_list:
                if hasattr(retriever, 'on_commit'):
                    handle_retriever_dependency(retriever, "commit", player_unit_retriever, host_uuid)

    def __repr__(self):
        return f"[RetrieverObjectLink] {self.name}: {self.section_name}.{self.link}" + \
               (f"\n\t- Process as: {self.process_as_object.__name__}" if self.process_as_object else "") + \
               (f"\n\t- Get Instance Number: True" if self.retrieve_instance_number else "") + \
               (f"\n\t- Get Hist Number: {self.retrieve_history_number}" if self.retrieve_history_number >= 0 else "")



import typing

_KT = typing.TypeVar("_KT")  # Key type
_VT = typing.TypeVar("_VT")  # Value type


class AttrDict(dict, typing.Mapping[_KT, _VT]):
    def __init__(self, *args):
        dict.__init__(self, args)

    def __getattribute__(self, item):
        try:
            return self[item]
        except (AttributeError, KeyError):
            try:
                return dict.__getattribute__(self, item)
            except AttributeError:
                return None

    def __setattr__(self, key, value):
        self[key] = value



class DependencyEval:
    def __init__(self, eval_code, eval_locals=None):
        """
        Object for storing dependency eval code and it's locals.

        Args:
            eval_code (str): The code executed using eval
            eval_locals (dict): The locals dict handed to the eval function
        """
        if eval_locals is None:
            eval_locals = {}

        self.eval_code = eval_code
        self.eval_locals = eval_locals

    @classmethod
    def instance_or_none(cls, eval_code):
        if eval_code is None:
            return None
        return cls(eval_code)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read().replace(":heavy_check_mark:", "✔️").replace(":x:", "❌")

with open("requirements.txt", "r") as r:
    dependencies = [line.strip() for line in r.readlines() if line.strip() and not line.strip().startswith("#")]

setuptools.setup(
    name="AoE2ScenarioParser",
    version="0.1.32",
    author="Kerwin Sneijders",
    author_email="ksneijders@hotmail.com",
    description="This is a project for editing parts of an 'aoe2scenario' file from Age of Empires 2 Definitive Edition",
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KSneijders/AoE2ScenarioParser",
    packages=setuptools.find_namespace_packages(),

    # According to the docs glob patterns are allowed.
    # But when using: "versions/**/*.json", it does not work.
    package_data={
        'AoE2ScenarioParser': [
            'datasets/sources/*.json',
            'versions/*/*/*.json'
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=dependencies
)



from typing import List, overload, Union, TYPE_CHECKING, Optional
from uuid import UUID

from AoE2ScenarioParser.datasets.players import PlayerId
from AoE2ScenarioParser.scenarios.scenario_store import store

if TYPE_CHECKING:
    from AoE2ScenarioParser.objects.data_objects.unit import Unit
    from AoE2ScenarioParser.objects.support.area import Area
    from AoE2ScenarioParser.objects.data_objects.trigger import Trigger


@overload
def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], unit: 'Unit') -> None:
    ...


@overload
def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], units: List['Unit']) -> None:
    ...


def unit_change_ownership(uuid: UUID, player: Union[int, PlayerId], *args) -> None:
    """
    Change the unit(s) ownership.

    Args:
        uuid: The UUID of the scenario
        player: The player to transfer the units to.
        args: Unit object or List of unit objects
    """

    def transfer_unit(scenario_, unit_, player_):
        scenario_.unit_manager.units[unit_.player].remove(unit_)
        scenario_.unit_manager.units[player_].append(unit_)

    scenario = store.get_scenario(uuid)
    if scenario:
        for units in args:
            if isinstance(units, List):
                for unit in units:
                    transfer_unit(scenario, unit, player)
            else:
                transfer_unit(scenario, units, player)


def import_triggers(uuid: UUID, triggers: List['Trigger'], insert_index: int = -1) -> None:
    """
    Import triggers into the scenario using the trigger manager.

    Args:
        uuid: The UUID of the scenario
        triggers: The trigger to import.
        insert_index: The insert index used in import triggers function
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        scenario.trigger_manager.import_triggers(triggers, insert_index)


def remove_triggers(uuid: UUID, trigger_ids: List[int]) -> None:
    """
    Import triggers into the scenario using the trigger manager.

    Args:
        uuid: The UUID of the scenario
        trigger_ids: A list of trigger ids to remove.
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        scenario.trigger_manager.triggers = [
            trigger for trigger in scenario.trigger_manager.triggers if trigger.trigger_id not in trigger_ids
        ]
        scenario.trigger_manager.reorder_triggers()


def new_area_object(uuid: UUID) -> Optional['Area']:
    """
    Creates a new area object.

    Args:
        uuid: The UUID of the scenario
    """
    scenario = store.get_scenario(uuid)
    if scenario:
        return scenario.new.area()



from typing import Union, Any, List


def add_str_trail(string: bytes) -> bytes:
    if len(string) > 0:
        string += (b"\x00" if string[-1] != 0 else "")
    else:
        return b"\x00"
    return string


def has_str_trail(string: bytes) -> bool:
    return len(string) > 0 and string[-1] == 0


def del_str_trail(string: Union[str, bytes]) -> Union[str, bytes]:
    if has_str_trail(string):
        string = string[:-1]
    return string


def add_prefix_chars(string: str, char: str, length: int):
    if len(string) > length:
        return string
    else:
        return char * (length - len(string)) + string


def add_suffix_chars(string: str, char: str, total_length: int):
    if len(string) > total_length:
        return string
    else:
        return string + char * (total_length - len(string))


def remove_prefix(string: str, prefix: str) -> str:
    """
    Cheap knockoff function of:
        https://docs.python.org/3/library/stdtypes.html?highlight=removesuffix#str.removeprefix

    Args:
        string (str): The string to check it's prefix
        prefix (str): The prefix to remove from the string

    Returns:
        The given string or the string without the prefix if it was present
    """
    return string[len(prefix):] if string.index(prefix) == 0 else string


def q_str(value: any) -> str:
    if type(value) is str:
        return f"'{value}'"
    if type(value) is bytes:
        return f"b'{value.hex()}'"
    return str(value)


def trunc_string(string: str, length=30, add_ellipsis=True):
    return (string[:length] + ('...' if add_ellipsis else '')) if len(string) > length else string


def trunc_bytes(string: bytes, length=30, add_ellipsis=True):
    return (string[:length] + (b'...' if add_ellipsis else b'')) if len(string) > length else string


def add_tabs(string: str, tabs: int) -> str:
    splitted_string = string.splitlines(keepends=True)
    return ("\t" * tabs) + ("\t" * tabs).join(splitted_string)


def create_inline_line(entries: List[Any]):
    return "\t" + ", ".join(map(str, entries)) + "\r\n"


def create_textual_hex(string: str, space_distance: int = 2, enter_distance: int = 48):
    """Please note that the 'enter_distance' parameter is including the - to be added - spaces. If you calculated it
    without the spaces, please multiply the number by: `block size incl space / block size excl space`"""
    return insert_char(insert_char(string, " ", space_distance), "\n", enter_distance)


# Credits: gurney alex @ https://stackoverflow.com/a/2657733/7230293
def insert_char(string: str, char: str, step=64):
    return char.join(string[i:i + step] for i in range(0, len(string), step))



from __future__ import annotations

from copy import deepcopy
from enum import Enum
from typing import Dict

from AoE2ScenarioParser.helper import bytes_parser
from AoE2ScenarioParser.helper.incremental_generator import IncrementalGenerator
from AoE2ScenarioParser.helper.list_functions import listify
from AoE2ScenarioParser.helper.pretty_format import pretty_format_list, pretty_format_dict
from AoE2ScenarioParser.helper.string_manipulations import create_textual_hex, insert_char, add_suffix_chars, q_str
from AoE2ScenarioParser.sections.aoe2_struct_model import AoE2StructModel, model_dict_from_structure
from AoE2ScenarioParser.sections.dependencies.dependency import handle_retriever_dependency
from AoE2ScenarioParser.sections.retrievers.retriever import Retriever, duplicate_retriever_map, reset_retriever_map


class SectionLevel(Enum):
    TOP_LEVEL = 0
    STRUCT = 1


class SectionName(Enum):
    FILE_HEADER = "FileHeader"
    DATA_HEADER = "DataHeader"
    MESSAGES = "Messages"
    CINEMATICS = "Cinematics"
    BACKGROUND_IMAGE = "BackgroundImage"
    PLAYER_DATA_TWO = "PlayerDataTwo"
    GLOBAL_VICTORY = "GlobalVictory"
    DIPLOMACY = "Diplomacy"
    OPTIONS = "Options"
    MAP = "Map"
    UNITS = "Units"
    TRIGGERS = "Triggers"
    FILES = "Files"


class AoE2FileSection:
    def __init__(self, name, retriever_map, host_uuid, struct_models=None, level=SectionLevel.TOP_LEVEL):
        if struct_models is None:
            struct_models = {}

        self.name: str = name
        self.retriever_map: Dict[str, 'Retriever'] = retriever_map
        self._host_uuid = host_uuid
        self.byte_length: int = -1
        self.struct_models: Dict[str, AoE2StructModel] = struct_models
        self.level: SectionLevel = level

    @classmethod
    def from_model(cls, model, host_uuid, set_defaults=False) -> AoE2FileSection:
        """
        Create a copy (what was called struct before) from a model.

        Args:
            model (AoE2StructModel): The model to copy from
            host_uuid (UUID): String representing host scenario
            set_defaults (bool): If retrievers need to be set to the default values

        Returns:
            An AoE2FileSection instance based on the model
        """
        duplicate_rmap = duplicate_retriever_map(model.retriever_map)
        if set_defaults:
            reset_retriever_map(duplicate_rmap)

        return cls(
            name=model.name,
            retriever_map=duplicate_rmap,
            host_uuid=host_uuid,
            struct_models=model.structs,
            level=SectionLevel.STRUCT
        )

    @classmethod
    def from_structure(cls, section_name, structure, host_uuid):
        retriever_map = {}
        for name, attr in structure.get('retrievers').items():
            retriever_map[name] = Retriever.from_structure(name, attr)

        structs = model_dict_from_structure(structure)
        return cls(section_name, retriever_map, host_uuid, structs)

    def get_data_as_bytes(self):
        result = []
        retriever: Retriever
        for retriever in self.retriever_map.values():
            result.append(retriever.get_data_as_bytes())
        return b''.join(result)

    def set_data_from_generator(self, igenerator: IncrementalGenerator) -> None:
        """
        Fill data from all retrievers with data from the given generator. Generator is expected to return bytes.
        Bytes will be parsed based on the retrievers. The total length of bytes read to fill this section is also stored
        in this section as `byte_length`.

        Args:
            igenerator: A generator from a binary scenario file
        """
        total_length = 0
        for retriever in self.retriever_map.values():
            handle_retriever_dependency(retriever, "construct", self, self._host_uuid)
            if retriever.datatype.type == "struct":
                struct_name = retriever.datatype.get_struct_name()

                retriever.data = []
                for _ in range(retriever.datatype.repeat):
                    model = self.struct_models.get(struct_name)
                    if model is None:
                        raise ValueError(f"Model '{struct_name}' not found. Likely not defined in structure.")

                    struct = self._create_struct(model, igenerator)
                    retriever.data.append(struct)

                    total_length += struct.byte_length
            else:
                retrieved_bytes = bytes_parser.retrieve_bytes(igenerator, retriever)
                self._fill_retriever_with_bytes(retriever, retrieved_bytes)

                total_length += sum([len(raw_bytes) for raw_bytes in retrieved_bytes])

        self.byte_length = total_length

    def _fill_retriever_with_bytes(self, retriever, retrieved_bytes):
        try:
            retriever.set_data_from_bytes(retrieved_bytes)
        except ValueError as e:
            print("\n\n" + "#" * 120)
            print(f"[{e.__class__.__name__}] Occurred while setting data in:\n\t{retriever}")
            print("#" * 120)
            print(self.get_byte_structure_as_string())
            raise e

    def _create_struct(self, model: AoE2StructModel, igenerator) -> AoE2FileSection:
        struct = AoE2FileSection.from_model(model, host_uuid=self._host_uuid)

        try:
            struct.set_data_from_generator(igenerator)
        except (TypeError, ValueError) as e:
            print("\n" + "#" * 120)
            print(f"[{e.__class__.__name__}] Occurred while setting data in: {struct.name}"
                  f"\n\tContent of {self.name}:")
            print("#" * 120)
            print(self.get_byte_structure_as_string())
            raise e

        return struct

    def set_data(self, data):
        retrievers = list(self.retriever_map.values())

        if len(data) == len(retrievers):
            for i in range(len(data)):
                retrievers[i].data = data[i]

                if hasattr(retrievers[i], 'on_construct'):
                    handle_retriever_dependency(retrievers[i], "construct", self, self._host_uuid)
        else:
            print(f"\nError in: {self.__class__.__name__}")
            print(f"Data: (len: {len(data)}) "
                  f"{pretty_format_list([f'{i}: {str(x)}' for i, x in enumerate(data)])}")
            print(f"Retrievers: (len: {len(retrievers)}) "
                  f"{pretty_format_list([f'{i}: {str(x)}' for i, x in enumerate(self.retriever_map.values())])}")
            raise ValueError("Data list isn't the same size as the DataType list")

    def __getattr__(self, item):
        """Providing a default way to access retriever data labeled 'name'"""
        if 'retriever_map' not in self.__dict__:
            return super().__getattribute__(item)
        elif item.startswith("__"):
            return super().__getattribute__(item)
        else:
            retriever = self.retriever_map[item]
            if retriever is None:
                return super().__getattribute__(item)
            else:
                return retriever.data

    def __setattr__(self, name, value):
        """Trying to edit retriever data labeled 'name' if available"""
        if 'retriever_map' not in self.__dict__:
            super().__setattr__(name, value)
        else:
            try:
                self.retriever_map[name].data = value
            except KeyError:
                super().__setattr__(name, value)

    def _entry_to_string(self, name, data, datatype):
        prefix = "\t"
        if self.level == SectionLevel.STRUCT:
            prefix = "\t\t\t"
        if 'str' in datatype:
            data = q_str(data)
        return f"{prefix}{name}: {data} ({datatype})\n"

    def get_header_string(self):
        if self.level == SectionLevel.TOP_LEVEL:
            return "######################## " + self.name + " ######################## [SECTION]"
        elif self.level == SectionLevel.STRUCT:
            return "############ " + self.name + " ############  [STRUCT]"

    def get_byte_structure_as_string(self):
        byte_structure = "\n" + self.get_header_string()

        for key, retriever in self.retriever_map.items():
            byte_structure += "\n"

            listed_retriever_data = listify(retriever.data)
            struct_header_set = False
            for struct in listed_retriever_data:
                if isinstance(struct, AoE2FileSection):
                    if not struct_header_set:
                        byte_structure += f"\n{'#' * 27} {key} ({retriever.datatype.to_simple_string()})"
                        struct_header_set = True
                    byte_structure += struct.get_byte_structure_as_string()
            # Struct Header was set. Retriever was struct, data retrieved using recursion. Next retriever.
            if struct_header_set:
                byte_structure += f"{'#' * 27} End of: {key} ({retriever.datatype.to_simple_string()})\n"
                continue

            retriever_data_bytes = retriever.get_data_as_bytes().hex()
            retriever_hex = create_textual_hex(retriever_data_bytes, space_distance=2, enter_distance=24)

            split_hex = retriever_hex.split("\n")
            split_hex_length = len(split_hex)

            retriever_short_string: str = retriever.get_short_str()
            split_data_string = retriever_short_string.replace('\x00', ' ').splitlines()
            data_lines = []
            for x in split_data_string:
                if len(x) > 120:
                    data_lines += [f'\t{x}' for x in insert_char(x, '\r\n', 120).splitlines()]
                else:
                    data_lines.append(x)
            split_data_length = len(data_lines)

            lines = max(split_hex_length, split_data_length)

            combined_strings = []
            for i in range(0, lines):
                hex_part = split_hex[i] if i < split_hex_length else ""
                data_part = data_lines[i] if i < split_data_length else ""
                combined_strings.append(add_suffix_chars(hex_part, " ", 28) + data_part)

            byte_structure += "\n".join(combined_strings)

        return byte_structure + "\n"

    def __str__(self):
        represent = self.name + ": \n"

        for retriever in self.retriever_map.values():
            if type(retriever.data) is list and len(retriever.data) > 0:
                if isinstance(retriever.data[0], AoE2FileSection):
                    represent += "\t" + retriever.name + ": [\n"
                    for x in retriever.data:
                        represent += "\t\t" + str(x)
                    represent += "\t]\n"
                else:
                    represent += self._entry_to_string(
                        retriever.name,
                        str(retriever.data),
                        str(retriever.datatype.to_simple_string())
                    )
            else:
                data = retriever.data or "None"
                represent += self._entry_to_string(
                    retriever.name, str(data), str(retriever.datatype.to_simple_string())
                )

        return represent

    def __repr__(self):
        return f"<AoE2FileSection> {self.name}"



from __future__ import annotations

import json
from enum import Enum
from pathlib import Path

with (Path(__file__).parent / 'sources' / f'unit_projectile_info.json').open() as file:
    # Open projectiles source-file for raw data. Used in `ProjectileInfo.get_unit_projectile(...)`
    _unit_projectile_info: dict = json.load(file)


class ProjectileInfo(Enum):
    """
    **Description**

    This enum class provides information about the projectiles in the game. Information about the following properties
    of a projectile is found in this class:
     - Projectile ID
     - The units that use the projectile

    **Methods**
    >>> ProjectileInfo.from_id()
    >>> ProjectileInfo.get_unit_projectile()

    **Examples**

    >>> ProjectileInfo.SCORPION_FIRE.ID
    >>> 628

    >>> ProjectileInfo.SCORPION_FIRE.USED_BY
    >>> (542,)
    """

    @property
    def ID(self) -> int:
        """
        Returns:
            The ID of the specified projectile unit
        """
        return self.value[0]

    @property
    def USED_BY(self) -> tuple[int]:
        """
        Returns:
            A tuple of unit IDs that use the specified projectile unit
        """
        return self.value[1]

    @classmethod
    def from_id(cls, projectile_id: int) -> ProjectileInfo:
        """
        Get the ProjectileInfo object from its ID

        Args:
            projectile_id: The ID of the projectile to get the ProjectileInfo of

        Returns:
            A ProjectileInfo object of the specified projectile ID
        """
        if type(projectile_id) is not int:
            raise TypeError(f"from_id expected int, got {type(projectile_id)}")
        if projectile_id == -1:
            raise ValueError("-1 is not a valid projectile ID")

        for projectile in cls._member_map_.values():
            if projectile.value[0] == projectile_id:
                return projectile

        raise KeyError(f"A projectile unit with ID '{projectile_id}' was not found in the dataset")

    @staticmethod
    def get_unit_projectile(unit_const: int, has_chemistry: bool = False, secondary: bool = False) -> ProjectileInfo:

        """
        Get the projectile object based on unit const

        Args:
            unit_const (int): The unit to find the projectile of
            has_chemistry (bool): If you want the fire (chemistry) version (if it exists)
            secondary (bool): If you want the secondary projectile version (if it exists)

        Returns:
            The ProjectileInfo corresponding to the given unit and params or None if nothing was found
        """

        if type(unit_const) is not int:
            raise TypeError(f"unit const expected int, got {type(unit_const)}")
        if unit_const < 0:
            raise ValueError(f"{unit_const} is not a valid unit constant")

        projectile = "secondary" if secondary else "primary"
        fire = "_fire" if has_chemistry else ""
        projectile_id = _unit_projectile_info.get(str(unit_const), {}).get(projectile + fire, -1)

        return ProjectileInfo.from_id(projectile_id) if projectile_id != -1 else None

    ARROW = 9, ()
    VOL = 54, (71, 109, 141, 142, 484, 597, 617, 621)
    ARROW1 = 97, ()
    SLINGER = 187, (185, 1075)
    CATST = 242, ()
    CATSF = 244, ()
    BOLTX = 245, ()
    BOLTF = 246, ()
    ARROW2 = 312, ()
    TREBST = 313, ()
    MANGST = 314, ()
    ARROW3 = 315, ()
    ARROW4 = 316, ()
    ARROW5 = 317, ()
    ARROW6 = 318, ()
    ARROW7 = 319, ()
    ARROW8 = 320, ()
    ARROW9 = 321, ()
    ARROW10 = 322, ()
    CATST1 = 323, ()
    CATST2 = 324, ()
    CATST3 = 325, ()
    CATST4 = 326, ()
    CATST5 = 327, ()
    VOL_FIRE = 328, (71, 109, 141, 142, 484, 597, 617, 621)
    ARW_FLM = 360, ()
    ARC = 363, (4, 158, 571)
    CROSSBOWMAN = 364, (24, 493, 866, 868, 873, 875, 926, 930, 1072, 1106, 1166, 1294, 1297)
    CRS = 365, (7,)
    HCS = 366, (6, 583, 596, 1010, 1012, 1036, 1079, 1155)
    SCORPION = 367, (279,)
    BOMBARD_CANNON = 368, (36, 644, 650, 1709)
    MANGONEL_SECONDARY = 369, (280, 550, 588)
    TREBUCHET = 371, (42, 682, 683, 1690)
    GAL = 372, (21,)
    WAR_GALLEY = 373, (442, 827, 829)
    CANNON_GALLEON = 374, (420, 691)
    COM_FIRE = 375, (24, 493, 866, 868, 873, 875, 926, 930, 1072, 1106, 1166, 1294, 1297)
    CRS_FIRE = 376, (7,)
    HCS_FIRE = 377, (6, 583, 596, 1010, 1012, 1036, 1079, 1155)
    SCORPION_FIRE = 378, (279,)
    GUNPOWDER_PRIMARY = 380, (5, 46, 52, 203, 425, 427, 557, 748, 771, 773, 1001, 1003, 1068)
    BOLTXF = 381, ()
    BOLTF1 = 385, ()
    MNG_FLM = 462, ()
    ARC_FIRE = 466, (4, 158, 571)
    MANGONEL_SECONDARY_FIRE = 468, (280, 550, 588)
    TREBUCHET_FIRE = 469, (42, 682, 683, 1690)
    GALLEY_FIRE = 470, (21,)
    WAR_GALLEY_FIRE = 471, (442, 827, 829)
    HAR_FIRE = 475, (11, 39, 172, 561, 577, 731, 1007, 1009, 1034, 1231, 1233, 1259, 1260, 1261, 1269, 1275, 1276)
    HHA_FIRE = 476, (437, 474, 698, 943, 1267, 1303)
    HAR = 477, (11, 39, 172, 561, 577, 731, 1007, 1009, 1034, 1231, 1233, 1259, 1260, 1261, 1269, 1275, 1276)
    HHA = 478, (437, 474, 698, 943, 1267, 1303)
    T_ARO = 485, ()
    WTN = 503, ()
    ARROW_GUARD_TOWER = 504, (79, 234, 235, 566, 785, 1665)
    KEP = 505, (79, 234, 235, 684, 685, 785, 1102, 1665)
    BTW = 506, (236,)
    ACA = 507, (492, 642, 686, 1182)
    AHX = 508, ()
    VIL = 509, (122, 216)
    CKN = 510, (73, 559, 1073, 1231, 1233, 1260)
    LONGBOWMAN = 511, (8, 200, 530, 763, 765, 850, 1069, 1129, 1131)
    LBT = 512, (45, 47, 51, 106, 133, 250, 533, 778, 885, 1189)
    MSU = 513, ()
    MPC = 514, ()
    TAX = 515, (165, 281, 424, 426, 531, 931, 1298)
    WTN_FIRE = 516, ()
    GTW_FIRE = 517, (79, 234, 235, 566, 785, 1665)
    KEP_FIRE = 518, (79, 234, 235, 684, 685, 785, 1102, 1665)
    ACA_FIRE = 519, (492, 642, 686, 1182)
    AHX_FIRE = 520, ()
    VIL_FIRE = 521, (122, 216)
    CKN_FIRE = 522, (73, 559, 1073, 1231, 1233, 1260)
    LBM_FIRE = 523, (8, 200, 530, 763, 765, 850, 1069, 1129, 1131)
    LBT_FIRE = 524, (45, 47, 51, 106, 133, 250, 533, 778, 885, 1189)
    MPC_FIRE = 525, ()
    MSU_FIRE = 526, ()
    FRG = 537, ()
    HFG = 538, ()
    SGY = 540, (539,)
    SGY_FIRE = 541, (539,)
    ONAGER = 551, ()
    ONAGER_FIRE = 552, ()
    HEAVY_SCORPION = 627, (542,)
    HEAVY_SCORPION_FIRE = 628, (542,)
    MANGONEL_PRIMARY = 656, (280, 550, 588)
    GP1 = 657, ()
    MANGONEL_PRIMARY_FIRE = 658, (280, 550, 588)
    FIRE_SHIP = 676, (188, 190, 529, 532, 938, 1103, 1302)
    MLK = 736, (282, 556, 929, 1296)
    CST = 746, (33, 82, 445, 1251)
    CST_FIRE = 747, (33, 82, 445, 1251)
    CGX = 767, (831, 832, 844, 1631)
    STW = 786, ()
    STW_FIRE = 787, ()
    KNIFE = 1055, (1013, 1015, 1066)
    CVB = 1057, (1004, 1006)
    CVB_FIRE = 1058, (1004, 1006)
    LBOL = 1111, ()
    LBOL_FIRE = 1112, ()
    HEAVY_SCORPION_UNUSED = 1113, ()
    HEAVY_SCORPION_FIRE_UNUSED = 1114, ()
    GUNPOWDER_SECONDARY = 1119, (203, 1001, 1003, 1733)
    BALLISTA_ELEPHANT = 1167, (1120, 1122)
    BALLISTA_ELEPHANT_FIRE = 1168, (1120, 1122)
    ARAMBAI = 1169, (1126, 1128)
    ARAMBAI_FIRE = 1170, (1126, 1128)
    SHCOW = 1223, ()
    LASER = 1595, (1222, 1577)
    HUSSITE_WAGON = 1733, (1704, 1706)



class InvalidScenarioStructureError(Exception):
    pass


class UnknownScenarioStructureError(Exception):
    pass


class UnknownStructureError(Exception):
    pass


class EndOfFileError(Exception):
    pass


class UnsupportedAttributeError(Exception):
    pass


class WarningToError(Exception):
    pass


def type_error_message(value, include_hint=True):
    return f"Expected int, found: {value.__class__}. " + (f"Maybe you meant: '{value}.ID'?" if include_hint else "")



from typing import List, Dict, Type, Tuple

from AoE2ScenarioParser.helper.string_manipulations import create_inline_line, add_tabs

_default_inline_types: List[Tuple[Type, int]] = [
    (int, 8),
    (float, 8)
]


def pretty_format_list(plist: List, inline_types: List[Tuple[Type, int]] = None):
    if len(plist) == 0:
        return "[]"
    if inline_types is None:
        inline_types = _default_inline_types

    return_string = "[\r\n"
    line_items = []
    for index, entry in enumerate(plist):
        for type_, n in inline_types:
            if isinstance(entry, type_):
                line_items.append(entry)
                if index % n == n - 1:
                    return_string += create_inline_line(line_items)
                    line_items = []
                break
        else:
            return_string += f"\t{entry}\r\n"
    if len(line_items) != 0:
        return_string += create_inline_line(line_items)
    return return_string + "]"


def pretty_format_dict(pdict: dict):
    return_string = "{\n"
    for key, value in pdict.items():
        if type(value) is dict:
            value = add_tabs(pretty_format_dict(value), 1)
        if type(value) is list:
            value = add_tabs(pretty_format_list(value), 1)
        newline = f"\t{key}: {value}"
        if newline[::-2] != "\n":
            newline += "\n"
        return_string += newline
    return return_string + "}\n"


def pretty_format_name(name: str) -> str:
    """
    Returns a pretty-printed version of the name string.
    Replaces all underscores with spaces and capitalizes the first letter
    of each word.
    For example, elite_chu_ko_nu -> Elite Chu Ko Nu.

    :Author:
        T-West (https://github.com/twestura/)
    """
    return ' '.join(s.capitalize() for s in name.split('_'))
